"""
Serializers for the dataclasses interface.

`pure-protobuf` contributors © 2011-2019
"""

import struct
from abc import ABC
from enum import IntEnum
from itertools import count
from struct import pack, unpack
from typing import Any, Type

from pure_protobuf.enums import WireType
from pure_protobuf.io_ import IO, Dumps, Loads


class Serializer(Dumps, Loads, ABC):
    # Default wire type for a value.
    # May be overridden by a wrapping serializer.
    wire_type: WireType

    def validate(self, value: Any):
        """
        Validates the value. Raises an exception if value incorrect.
        """

    def merge(self, old_value: Any, new_value: Any) -> Any:
        """
        Merges two unrepeated values on a wire.
        """
        return new_value


class UnsignedVarintSerializer(Serializer):
    """
    Serializes a generic base-128 unsigned varint.
    See also: https://developers.google.com/protocol-buffers/docs/encoding#varints
    See also: https://en.wikipedia.org/wiki/LEB128
    """

    wire_type = WireType.VARINT

    def validate(self, value: Any):
        if not isinstance(value, int):
            raise ValueError('an integer is expected')

    def dump(self, value: Any, io: IO):
        while value > 0x7F:
            io.write(bytes((value & 0x7F | 0x80,)))
            value >>= 7
        io.write(bytes((value,)))

    def load(self, io: IO) -> Any:
        value = 0
        for shift in count(0, 7):
            byte, = io.read(1)
            value |= (byte & 0x7F) << shift
            if not byte & 0x80:
                return value


unsigned_varint_serializer = UnsignedVarintSerializer()


class SignedVarintSerializer(Serializer):
    """
    Serializes a generic base-128 signed varint.
    See also: https://developers.google.com/protocol-buffers/docs/encoding#signed-integers
    """

    wire_type = unsigned_varint_serializer.wire_type

    def validate(self, value: Any):
        if not isinstance(value, int):
            raise ValueError('an integer is expected')

    def dump(self, value: Any, io: IO):
        return unsigned_varint_serializer.dump(abs(value) * 2 - (value < 0), io)

    def load(self, io: IO) -> Any:
        # See also: https://stackoverflow.com/a/2211086/359730
        n = unsigned_varint_serializer.load(io)
        return (n >> 1) ^ (-(n & 1))


signed_varint_serializer = SignedVarintSerializer()


class BytesSerializer(Serializer):
    """
    Serializes a byte string.
    See also: https://developers.google.com/protocol-buffers/docs/encoding#strings
    """

    wire_type = WireType.BYTES

    def validate(self, value: Any):
        if not isinstance(value, bytes):
            raise ValueError('a byte string is expected')

    def dump(self, value: Any, io: IO):
        unsigned_varint_serializer.dump(len(value), io)
        io.write(value)

    def load(self, io: IO) -> Any:
        length = unsigned_varint_serializer.load(io)
        return io.read(length)


bytes_serializer = BytesSerializer()


class StringSerializer(Serializer):
    """
    Serializes a UTF8-encoded string.
    """

    wire_type = bytes_serializer.wire_type

    def validate(self, value: Any):
        if not isinstance(value, str):
            raise ValueError('a string is expected')

    def dump(self, value: Any, io: IO):
        bytes_serializer.dump(value.encode('utf-8'), io)

    def load(self, io: IO) -> Any:
        return bytes_serializer.load(io).decode('utf-8')


class UnsignedInt32Serializer(Serializer):
    """
    Serializes 32-bit unsigned varint.
    """

    wire_type = unsigned_varint_serializer.wire_type

    def validate(self, value: Any):
        unsigned_varint_serializer.validate(value)
        if not 0 <= value <= 0xFFFFFFFF:
            raise ValueError(f'value is out of 32-bit unsigned integer range')

    def dump(self, value: Any, io: IO):
        unsigned_varint_serializer.dump(value, io)

    def load(self, io: IO) -> Any:
        return unsigned_varint_serializer.load(io)


class UnsignedInt64Serializer(Serializer):
    """
    Serializes 64-bit unsigned varint.
    """

    wire_type = unsigned_varint_serializer.wire_type

    def validate(self, value: Any):
        unsigned_varint_serializer.validate(value)
        if not 0 <= value <= 0xFFFFFFFF_FFFFFFFF:
            raise ValueError(f'value is out of 64-bit unsigned integer range')

    def dump(self, value: Any, io: IO):
        unsigned_varint_serializer.dump(value, io)

    def load(self, io: IO) -> Any:
        return unsigned_varint_serializer.load(io)


class SignedInt32Serializer(Serializer):
    """
    Serializes 32-bit signed varint.
    """

    wire_type = signed_varint_serializer.wire_type

    def validate(self, value: Any):
        signed_varint_serializer.validate(value)
        if not -0x7FFFFFFF <= value <= 0x7FFFFFFF:
            raise ValueError(f'value is out of 32-bit signed integer range')

    def dump(self, value: Any, io: IO):
        signed_varint_serializer.dump(value, io)

    def load(self, io: IO) -> Any:
        return signed_varint_serializer.load(io)


class SignedInt64Serializer(Serializer):
    """
    Serializes 64-bit signed varint.
    """

    wire_type = signed_varint_serializer.wire_type

    def validate(self, value: Any):
        signed_varint_serializer.validate(value)
        if not -0x7FFFFFFF_FFFFFFFF <= value <= 0x7FFFFFFF_FFFFFFFF:
            raise ValueError(f'value is out of 64-bit signed integer range')

    def dump(self, value: Any, io: IO):
        signed_varint_serializer.dump(value, io)

    def load(self, io: IO) -> Any:
        return signed_varint_serializer.load(io)


class BooleanSerializer(Serializer):
    """
    Serializes a boolean value.
    See also: https://developers.google.com/protocol-buffers/docs/proto3#scalar
    """

    wire_type = WireType.VARINT

    def validate(self, value: Any):
        if not isinstance(value, bool):
            raise ValueError('a boolean is expected')

    def dump(self, value: Any, io: IO):
        io.write(b'\x01' if value else b'\x00')

    def load(self, io: IO) -> Any:
        return bool(unsigned_varint_serializer.load(io))


class SignedFixed32Serializer(Serializer):
    """
    Serializes 32-bit signed integer.
    See also: https://developers.google.com/protocol-buffers/docs/encoding#non-varint-numbers
    """

    wire_type = WireType.LONG

    def validate(self, value: Any):
        if not isinstance(value, int):
            raise ValueError('an integer is expected')

    def dump(self, value: Any, io: IO):
        io.write(pack('<i', value))

    def load(self, io: IO) -> Any:
        return unpack('<i', io.read(4))[0]


class UnsignedFixed32Serializer(Serializer):
    """
    Serializes 32-bit unsigned integer.
    See also: https://developers.google.com/protocol-buffers/docs/encoding#non-varint-numbers
    """

    wire_type = WireType.LONG

    def validate(self, value: Any):
        if not isinstance(value, int):
            raise ValueError('an integer is expected')

    def dump(self, value: Any, io: IO):
        io.write(pack('<I', value))

    def load(self, io: IO) -> Any:
        return unpack('<I', io.read(4))[0]


class SignedFixed64Serializer(Serializer):
    """
    Serializes 64-bit signed integer.
    See also: https://developers.google.com/protocol-buffers/docs/encoding#non-varint-numbers
    """

    wire_type = WireType.LONG_LONG

    def validate(self, value: Any):
        if not isinstance(value, int):
            raise ValueError('an integer is expected')

    def dump(self, value: Any, io: IO):
        io.write(pack('<q', value))

    def load(self, io: IO) -> Any:
        return unpack('<q', io.read(8))[0]


class UnsignedFixed64Serializer(Serializer):
    """
    Serializes 64-bit unsigned integer.
    See also: https://developers.google.com/protocol-buffers/docs/encoding#non-varint-numbers
    """

    wire_type = WireType.LONG_LONG

    def validate(self, value: Any):
        if not isinstance(value, int):
            raise ValueError('an integer is expected')

    def dump(self, value: Any, io: IO):
        io.write(pack('<Q', value))

    def load(self, io: IO) -> Any:
        return unpack('<Q', io.read(8))[0]


class FloatSerializer(Serializer):
    """
    Serializes 32-bit floating-point value.
    See also: https://developers.google.com/protocol-buffers/docs/encoding#non-varint-numbers
    """

    wire_type = WireType.LONG

    def validate(self, value: Any):
        if not isinstance(value, float):
            raise ValueError('a floating-point value is expected')

    def dump(self, value: Any, io: IO):
        io.write(pack('<f', value))

    def load(self, io: IO) -> Any:
        return unpack('<f', io.read(4))[0]


class DoubleSerializer(Serializer):
    """
    Serializes 64-bit double floating-point value.
    See also: https://developers.google.com/protocol-buffers/docs/encoding#non-varint-numbers
    """

    wire_type = WireType.LONG_LONG

    def validate(self, value: Any):
        if not isinstance(value, float):
            raise ValueError('a floating-point value is expected')

    def dump(self, value: Any, io: IO):
        io.write(pack('<d', value))

    def load(self, io: IO) -> Any:
        return unpack('<d', io.read(8))[0]


class IntEnumSerializer(Serializer):
    """
    Serializes integer enumeration value.
    See also: https://developers.google.com/protocol-buffers/docs/proto3#enum
    """

    wire_type = WireType.VARINT

    def __init__(self, type_: Type[IntEnum]):
        self.type_ = type_

    def validate(self, value: Any):
        if not isinstance(value, self.type_):
            raise ValueError(f'{self.type_} instance is expected, got {type(value)}')

    def dump(self, value: Any, io: IO):
        unsigned_varint_serializer.dump(value, io)

    def load(self, io: IO):
        return self.type_(unsigned_varint_serializer.load(io))


class MessageSerializer(Serializer):
    """
    Serializes an entire message.
    """

    wire_type = WireType.BYTES

    def __init__(self, type_: Any):
        self.type_ = type_

    def validate(self, value: Any):
        if not isinstance(value, self.type_):
            raise ValueError(f'{self.type_} is expected, but got {type(value)}')
        for field_ in value.__protobuf_fields__.values():
            field_.validate(getattr(value, field_.name))

    def dump(self, value: Any, io: IO):
        for number, field_ in value.__protobuf_fields__.items():
            field_value = getattr(value, field_.name)
            try:
                field_.dump(field_value, io)
            except (ValueError, struct.error) as e:
                raise ValueError(f'field `{field_.name}`: {e}, got `{field_value}`') from e

    def load(self, io: IO) -> Any:
        values = {}
        while True:
            try:
                key = unsigned_varint_serializer.load(io)
            except ValueError:
                break
            wire_type = WireType(key & 0b111)
            field = self.type_.__protobuf_fields__.get(key >> 3)
            if field is not None:
                old_value = values.get(field.name)
                new_value = field.load(wire_type, io)
                values[field.name] = field.merge(old_value, new_value)
            else:
                SKIP[wire_type](io)
        return self.type_(**values)

    def merge(self, old_value: Any, new_value: Any) -> Any:
        # For embedded message fields, the parser merges multiple instances of the same field,
        # as if with the ``Message::MergeFrom`` method.
        # See also: https://developers.google.com/protocol-buffers/docs/encoding#optional
        if old_value is not None:
            old_value.merge_from(new_value)
            return old_value
        else:
            return new_value


class PackingSerializer(Serializer):
    """
    Wraps an inner serializer into a byte string serializer.
    """

    wire_type = bytes_serializer.wire_type

    def __init__(self, inner: Serializer):
        self.inner = inner

    def validate(self, value: Any):
        self.inner.validate(value)

    def dump(self, value: Any, io: IO):
        bytes_serializer.dump(self.inner.dumps(value), io)

    def load(self, io: IO) -> Any:
        return self.inner.loads(bytes_serializer.load(io))

    def merge(self, old_value: Any, new_value: Any) -> Any:
        return self.inner.merge(old_value, new_value)


def skip_varint(io: IO):
    while io.read(1)[0] & 0x80:
        pass


def skip_fixed_32(io: IO):
    io.read(4)


def skip_fixed_64(io: IO):
    io.read(8)


def skip_bytes(io: IO):
    io.read(unsigned_varint_serializer.load(io))


# Functions to skip unknown field values.
SKIP = {
    WireType.BYTES: skip_bytes,
    WireType.LONG: skip_fixed_32,
    WireType.LONG_LONG: skip_fixed_64,
    WireType.VARINT: skip_varint,
}



"""
`pure-protobuf` contributors © 2011-2019
"""

from abc import ABC, abstractmethod
from io import BytesIO
from typing import Any, Iterable

from pure_protobuf.enums import WireType
from pure_protobuf.io_ import IO, Dumps
from pure_protobuf.serializers import Serializer, bytes_serializer, unsigned_varint_serializer


class Field(Dumps, ABC):
    def __init__(self, number: int, name: str, serializer: Serializer):
        self.number = number
        self.name = name
        self.serializer = serializer
        self.wire_type = serializer.wire_type

    @abstractmethod
    def validate(self, value: Any):
        """
        Validates a value. Raises an exception in case of an error.
        """
        raise NotImplementedError()

    @abstractmethod
    def dump(self, value: Any, io: IO):
        """
        Serializes a value into a file-like object.
        """
        raise NotImplementedError()

    def dump_key(self, io: IO):
        """
        Serializes field key: number and wire type.
        """
        unsigned_varint_serializer.dump((self.number << 3) | self.wire_type, io)

    @abstractmethod
    def load(self, wire_type: WireType, io: IO) -> Any:
        """
        Deserializes a field value from a file-like object.
        Accepts received wire type.
        """
        raise NotImplementedError()

    @abstractmethod
    def merge(self, old_value: Any, new_value: Any) -> Any:
        """
        Merge existing value with a new value received from an input stream.
        """
        raise NotImplementedError()


class NonRepeatedField(Field):
    """
    Handles non-repeated field.
    See also: https://developers.google.com/protocol-buffers/docs/proto3#scalar
    See also: https://developers.google.com/protocol-buffers/docs/encoding#optional
    """

    def __init__(self, number: int, name: str, serializer: Serializer, is_optional: bool):
        super().__init__(number, name, serializer)
        self.is_optional = is_optional

    def validate(self, value: Any):
        if value is not None or not self.is_optional:
            self.serializer.validate(value)

    def dump(self, value: Any, io: IO):
        if value is not None:
            self.dump_key(io)
            self.serializer.dump(value, io)

    def load(self, wire_type: WireType, io: IO) -> Any:
        if wire_type != self.serializer.wire_type:
            raise ValueError(f'expected {self.serializer.wire_type}, got {wire_type}')
        return self.serializer.load(io)

    def merge(self, old_value: Any, new_value: Any) -> Any:
        return self.serializer.merge(old_value, new_value)


class RepeatedField(Field, ABC):
    """
    Handles repeated and packed repeated field.
    See also: https://developers.google.com/protocol-buffers/docs/encoding#optional
    """

    def validate(self, value: Any):
        for item in value:
            self.serializer.validate(item)

    def load(self, wire_type: WireType, io: IO) -> Any:
        if self.serializer.wire_type != WireType.BYTES and wire_type == WireType.BYTES:
            # Protocol buffer parsers must be able to parse repeated fields
            # that were compiled as packed as if they were not packed, and vice versa.
            # See also: https://developers.google.com/protocol-buffers/docs/encoding#packed
            return list(self.load_packed(bytes_serializer.load(io)))
        if wire_type == self.serializer.wire_type:
            return [self.serializer.load(io)]
        raise ValueError(f'expected {self.serializer.wire_type} or {WireType.BYTES}, got {wire_type}')

    def load_packed(self, bytes_: bytes) -> Iterable[Any]:
        """
        Deserializes repeated values from a packed byte string.
        """
        with BytesIO(bytes_) as io:
            while io.tell() < len(bytes_):
                yield self.serializer.load(io)

    def merge(self, old_value: Any, new_value: Any) -> Any:
        if old_value is not None:
            return [*old_value, *new_value]
        return new_value


class UnpackedRepeatedField(RepeatedField):
    """
    See also: https://developers.google.com/protocol-buffers/docs/encoding#optional
    """

    def dump(self, value: Any, io: IO):
        for item in value:
            self.dump_key(io)
            self.serializer.dump(item, io)


class PackedRepeatedField(RepeatedField):
    """
    Dumps repeated value as a byte string.
    See also: https://developers.google.com/protocol-buffers/docs/encoding#packed
    """

    def __init__(self, number: int, name: str, serializer: Serializer):
        super().__init__(number, name, serializer)
        self.wire_type = WireType.BYTES

    def dump(self, value: Any, io: IO):
        with BytesIO() as inner_io:
            for item in value:
                self.serializer.dump(item, inner_io)
            self.dump_key(io)
            bytes_serializer.dump(inner_io.getvalue(), io)



# coding: utf-8

"""
Legacy interface.

`pure-protobuf` contributors © 2011-2019
"""

from __future__ import absolute_import

import struct
import warnings
from io import BytesIO

from pure_protobuf.six import b

warnings.warn(
    'the legacy interface is deprecated and will be removed, use the dataclasses interface',
    DeprecationWarning,
)


class Type(object):
    """
    Represents a general field type.
    """

    def dump(self, fp, value):
        """
        Dumps its value to write-like object.
        """
        raise TypeError('Don\'t call this directly.')

    def load(self, fp):
        """
        Loads its value from read-like object and returns a read value.
        """
        raise TypeError('Don\'t call this directly.')

    def dumps(self, value):
        """
        Dumps its value to string and returns this string.
        """
        fp = BytesIO()
        self.dump(fp, value)
        return fp.getvalue()

    def loads(self, s):
        """
        Loads its value from a string and returns a read value.
        """
        return self.load(BytesIO(s))

    def __hash__(self):
        """
        Returns a hash of this type.
        """
        return hash(self.__class__.__name__)


class UVarintType(Type):
    """
    Represents an unsigned Varint type.
    """

    WIRE_TYPE = 0

    def dump(self, fp, value):
        shifted_value = True
        while shifted_value:
            shifted_value = value >> 7
            fp.write(bytearray(((value & 0x7F) | (0x80 if shifted_value != 0 else 0x00),)))
            value = shifted_value

    def load(self, fp):
        value, shift, quantum = 0, 0, 0x80
        while (quantum & 0x80) == 0x80:
            quantum = ord(fp.read(1))
            value, shift = value + ((quantum & 0x7F) << shift), shift + 7
        return value


class VarintType(UVarintType):
    """
    Represents a signed Varint type. Implements ZigZag encoding.
    """

    def dump(self, fp, value):
        encoded_varint = abs(value) << 1
        if value < 0:
            encoded_varint -= 1
        UVarintType.dump(self, fp, encoded_varint)

    def load(self, fp):
        encoded_varint = UVarintType.load(self, fp) + 1
        div = encoded_varint >> 1
        return div if encoded_varint & 1 else -div


class BoolType(UVarintType):
    """
    Represents a boolean type. Encodes True as UVarint 1, and False as
    UVarint 0.
    """

    def dump(self, fp, value):
        fp.write(b('\x01' if value else '\x00'))  # similarly to UVarint

    def load(self, fp):
        return UVarintType.load(self, fp) != 0


class BytesType(Type):
    """
    Represents a raw bytes type.
    """

    WIRE_TYPE = 2

    def dump(self, fp, value):
        UVarint.dump(fp, len(value))
        fp.write(value)

    def load(self, fp):
        return fp.read(UVarint.load(fp))


class UnicodeType(BytesType):

    def dump(self, fp, value):
        return BytesType.dump(self, fp, value.encode('utf-8'))

    def load(self, fp):
        return BytesType.load(self, fp).decode('utf-8')


class FixedLengthType(Type):
    """
    Represents a general fixed-length value type. You should not use this type
    directly. Use derived types instead.
    """

    def dump(self, fp, value):
        fp.write(value)

    def load(self, fp):
        return fp.read(self.length())

    def length(self):
        raise NotImplementedError()


class Fixed64Type(FixedLengthType):
    """
    Represents a general 64-bit value type.
    """

    WIRE_TYPE = 1

    def length(self):
        return 8


class Fixed32Type(FixedLengthType):
    """
    Represents a general 32-bit value type.
    """

    WIRE_TYPE = 5

    def length(self):
        return 4


class Fixed64SubType(Fixed64Type):
    """
    Represents a general pickle'able 64-bit value type.
    """

    dump = lambda self, fp, value: Fixed64Type.dump(
        self, fp, struct.pack(self.format, value)
    )

    load = lambda self, fp: struct.unpack(
        self.format, Fixed64Type.load(self, fp)
    )[0]


class UInt64Type(Fixed64SubType):
    """
    Represents an unsigned int64 type.
    """

    format = '<Q'


class Int64Type(Fixed64SubType):
    """
    Represents a signed int64 type.
    """

    format = '<q'


class Float64Type(Fixed64SubType):
    """
    Represents a double precision floating point type.
    """

    format = '<d'


class Fixed32SubType(Fixed32Type):
    """
    Represents a pickle'able 32-bit value.
    """

    dump = lambda self, fp, value: Fixed32Type.dump(
        self, fp, struct.pack(self.format, value)
    )

    load = lambda self, fp: struct.unpack(
        self.format, Fixed32Type.load(self, fp)
    )[0]


class UInt32Type(Fixed32SubType):
    """
    Represents an unsigned int32 type.
    """

    format = '<I'


class Int32Type(Fixed32SubType):
    """
    Represents a signed int32 type.
    """

    format = '<i'


class Float32Type(Fixed32SubType):
    """
    Represents a single precision floating point type.
    """

    format = '<f'


# Types instances. ------------------------------------------------------------
# You should use these types instances when defining your message type.

UVarint = UVarintType()
Varint = VarintType()
Bool = BoolType()
Fixed64 = Fixed64Type()
UInt64 = UInt64Type()
Int64 = Int64Type()
Float64 = Float64Type()
Fixed32 = Fixed32Type()
UInt32 = UInt32Type()
Int32 = Int32Type()
Float32 = Float32Type()
Bytes = BytesType()
Unicode = UnicodeType()


# Messages. -------------------------------------------------------------------

class Flags(object):
    """
    Flags for a field.
    """

    SIMPLE = 0  # Single value field.
    REQUIRED, REQUIRED_MASK = 1, 1  # Required field_type.
    # Repeated and packed-repeated fields.
    SINGLE, REPEATED, PACKED_REPEATED, REPEATED_MASK = (
        0,
        2,
        6,
        6,
    )
    # Used by MessageMetaType to determine if a field contains embedded
    # definition.
    PRIMITIVE, EMBEDDED, EMBEDDED_MASK = (
        0,
        8,
        8,
    )


class EofWrapper:
    """
    Wraps a stream to raise EOFError instead of just returning of ''.
    """

    def __init__(self, fp, limit=None):
        self.__fp = fp
        self.__limit = limit

    def read(self, size=None):
        """
        Reads a string. Raises EOFError on end of stream.
        """
        if size == 0:
            return ''
        if self.__limit is not None:
            size = min(size, self.__limit)
            self.__limit -= size
        s = self.__fp.read(size)
        if len(s) == 0:
            raise EOFError()
        return s


def _pack_key(tag, wire_type):
    """
    Packs a tag and a wire_type into single int according to the protobuf spec.
    """
    return (tag << 3) | wire_type


def _unpack_key(key):
    """
    Unpacks a key into a tag and a wire_type according to the protobuf spec.
    """
    return key >> 3, key & 7


# This used to correctly determine the length of unknown tags when loading a
# message.
_wire_type_to_type_instance = {0: Varint, 1: Fixed64, 2: Bytes, 5: Fixed32}


class MessageType(Type):
    """
    Represents a message type.
    """

    def __init__(self):
        """
        Creates a new message type.
        """
        self.__tags_to_types = dict()  # Maps a tag to a type instance.
        self.__tags_to_names = dict()  # Maps a tag to a given field name.
        self.__flags = dict()  # Maps a tag to flags.

    def __hash__(self):
        _hash = 17
        for tag, name, field_type, flags in iter(self):
            _hash = hash((_hash, tag, field_type, flags))
        return _hash

    def __iter__(self):
        """
        Iterates over all fields.
        """
        for tag, name in self.__tags_to_names.items():
            yield (tag, name, self.__tags_to_types[tag], self.__flags[tag])

    def add_field(self, tag, name, field_type, flags=Flags.SIMPLE):
        """
        Adds a field to the message type.
        """
        if tag in self.__tags_to_names or tag in self.__tags_to_types:
            raise ValueError('The tag %s is already used.' % tag)
        self.__tags_to_names[tag] = name
        self.__tags_to_types[tag] = field_type
        self.__flags[tag] = flags
        return self  # Allow add_field chaining.

    def remove_field(self, tag):
        """
        Removes a field by its tag. Doesn't raise any exception when the tag is
        missing.
        """
        if tag in self.__tags_to_names:
            del self.__tags_to_names[tag]
        if tag in self.__tags_to_types:
            del self.__tags_to_types[tag]

    def __call__(self):
        """
        Creates an instance of this message type.
        """
        return Message(self)

    def __has_flag(self, tag, flag, mask):
        """
        Checks whether the field with the specified tag has the specified flag.
        """
        return (self.__flags[tag] & mask) == flag

    def dump(self, fp, value):
        if self != value.message_type:
            raise TypeError(
                'Attempting to dump an object with type that\'s different '
                'from mine.'
            )
        for tag, field_type in self.__tags_to_types.items():
            if self.__tags_to_names[tag] in value:
                if self.__has_flag(tag, Flags.SINGLE, Flags.REPEATED_MASK):
                    # Single value.
                    UVarint.dump(fp, _pack_key(tag, field_type.WIRE_TYPE))
                    field_type.dump(fp, value[self.__tags_to_names[tag]])
                elif self.__has_flag(
                    tag, Flags.PACKED_REPEATED, Flags.REPEATED_MASK
                ):
                    # Repeated packed value.
                    UVarint.dump(fp, _pack_key(tag, Bytes.WIRE_TYPE))
                    internal_fp = BytesIO()
                    for single_value in value[self.__tags_to_names[tag]]:
                        field_type.dump(internal_fp, single_value)
                    Bytes.dump(fp, internal_fp.getvalue())
                elif self.__has_flag(tag, Flags.REPEATED, Flags.REPEATED_MASK):
                    # Repeated value.
                    key = _pack_key(tag, field_type.WIRE_TYPE)
                    # Put it together sequently.
                    for single_value in value[self.__tags_to_names[tag]]:
                        UVarint.dump(fp, key)
                        field_type.dump(fp, single_value)
            elif self.__has_flag(tag, Flags.REQUIRED, Flags.REQUIRED_MASK):
                raise ValueError(
                    'The field with the tag %s is required but a value is '
                    'missing.' % tag
                )

    def load(self, fp):
        fp, message = (
            EofWrapper(fp),
            self.__call__(),
        )  # Wrap fp and create a new instance.
        while True:
            try:
                tag, wire_type = _unpack_key(UVarint.load(fp))
                if tag in self.__tags_to_types:
                    field_type = self.__tags_to_types[tag]
                    if not self.__has_flag(
                        tag, Flags.PACKED_REPEATED, Flags.REPEATED_MASK
                    ):
                        if wire_type != field_type.WIRE_TYPE:
                            raise TypeError(
                                'The received value with the tag %s has '
                                'incorrect wiretype: %s instead of %s '
                                'expected.'
                                % (tag, wire_type, field_type.WIRE_TYPE)
                            )
                    elif wire_type != Bytes.WIRE_TYPE:
                        raise TypeError(
                            'Tag %s has wiretype %s while the field is packed '
                            'repeated.'
                            % (tag, wire_type)
                        )
                    if self.__has_flag(tag, Flags.SINGLE, Flags.REPEATED_MASK):
                        # Single value.
                        message[self.__tags_to_names[tag]] = field_type.load(
                            fp
                        )
                    elif self.__has_flag(
                        tag, Flags.PACKED_REPEATED, Flags.REPEATED_MASK
                    ):
                        # Repeated packed value.
                        repeated_value = message[
                            self.__tags_to_names[tag]
                        ] = list()
                        internal_fp = EofWrapper(
                            fp, UVarint.load(fp)
                        )  # Limit with value length.
                        while True:
                            try:
                                repeated_value.append(
                                    field_type.load(internal_fp)
                                )
                            except EOFError:
                                break
                    elif self.__has_flag(
                        tag, Flags.REPEATED, Flags.REPEATED_MASK
                    ):
                        # Repeated value.
                        if not self.__tags_to_names[tag] in message:
                            repeated_value = message[
                                self.__tags_to_names[tag]
                            ] = list()
                        repeated_value.append(field_type.load(fp))
                else:
                    # Skip this field.
                    _wire_type_to_type_instance[wire_type].load(fp)
            except EOFError:
                # Check if all required fields are present.
                for tag, name in self.__tags_to_names.items():
                    has_flag = self.__has_flag(
                        tag, Flags.REQUIRED, Flags.REQUIRED_MASK
                    )
                    if has_flag and (name not in message):
                        if self.__has_flag(
                            tag, Flags.REPEATED, Flags.REPEATED_MASK
                        ):
                            # Empty list (no values was in input stream).
                            # But required field.
                            message[name] = list()
                        else:
                            raise ValueError(
                                'The field with the tag %s (\'%s\') is '
                                'required but a value is missing.'
                                % (tag, name)
                            )
                return message


class Message(dict):
    """
    Represents a message instance.
    """

    def __init__(self, message_type):
        """
        Initializes a new instance of the specified message type.
        """
        super(Message, self).__init__()
        self.__dict__['message_type'] = message_type

    def __getattr__(self, name):
        """
        Gets a value of the specified message field.
        """
        return self.__getitem__(name)

    def __setattr__(self, name, value):
        """
        Sets a value of the specified message field.
        """
        mapping = self.__dict__ if name in self.__dict__ else self
        mapping.__setitem__(name, value)

        return value

    def dumps(self):
        """
        Dumps the message into a string.
        """
        return self.message_type.dumps(self)

    def dump(self, fp):
        """
        Dumps the message into a write-like object.
        """
        return self.message_type.dump(fp, self)


def loads(self, s, message_type):
    """
    Loads a message of the specified message type from the string.
    """
    return message_type.loads(s)


def load(self, fp, message_type):
    """
    Loads a message of the specified message type from the read-like object.
    """
    return message_type.load(fp)


# Embedded message. -----------------------------------------------------------

class EmbeddedMessage(Type):
    """
    Represents an embedded message type.
    """

    WIRE_TYPE = 2

    def __init__(self, message_type):
        """
        Initializes a new instance. The argument is an underlying message type.
        """
        self.message_type = message_type

    def __call__(self):
        """
        Creates a message of the underlying message type.
        """
        return self.message_type()

    def dump(self, fp, value):
        Bytes.dump(fp, self.message_type.dumps(value))

    def load(self, fp):
        return self.message_type.load(
            EofWrapper(fp, UVarint.load(fp))
        )  # Limit with embedded message length.



"""
Google Protocol Buffers well-known types.

`pure-protobuf` contributors © 2011-2019
"""

# noinspection PyCompatibility
from dataclasses import dataclass
from datetime import datetime, timezone
from math import modf
from typing import Any, Dict, Type

from pure_protobuf.dataclasses_ import field, message
from pure_protobuf.io_ import IO
from pure_protobuf.serializers import MessageSerializer, PackingSerializer, Serializer
from pure_protobuf.types import int32, int64


@message
@dataclass
class Timestamp:
    """
    See also: https://github.com/protocolbuffers/protobuf/blob/master/src/google/protobuf/timestamp.proto
    """
    seconds: int64 = field(1)
    nanos: int32 = field(2)


# noinspection PyPep8Naming
@message
@dataclass
class Any_:
    """
    See also: https://github.com/protocolbuffers/protobuf/blob/master/src/google/protobuf/any.proto
    """
    type_url: str = field(1, default='')
    value: bytes = field(2, default=b'')


class DateTimeSerializer(MessageSerializer):
    """
    Supports ``datetime`` as ``Timestamp`` well-known type.
    """

    def __init__(self):
        super().__init__(Timestamp)

    def validate(self, value: Any):
        if not isinstance(value, datetime):
            raise ValueError(f'`datetime` expected, got `{type(value)}`')

    def dump(self, value: Any, io: IO):
        fraction, whole = modf(value.timestamp())
        super().dump(Timestamp(seconds=int64(int(whole)), nanos=int32(int(fraction * 1_000_000_000.0))), io)

    def load(self, io: IO) -> Any:
        timestamp: Timestamp = super().load(io)
        return datetime.fromtimestamp(
            float(timestamp.seconds) + float(timestamp.nanos) / 1_000_000_000.0,
            tz=timezone.utc,
        )


SERIALIZERS: Dict[Type, Serializer] = {
    # TODO: Any
    datetime: PackingSerializer(DateTimeSerializer()),
}



#!/usr/bin/env python3
# coding: utf-8

"""
Tiny Python 2 & 3 compatibility layer.

`pure-protobuf` contributors © 2011-2019
"""

from __future__ import absolute_import

import sys

PYTHON_3 = sys.version_info.major == 3

if PYTHON_3:
    def b(string):
        return bytes(string, encoding='latin=1')
else:
    def b(string):
        return string



"""
Type hints for the dataclasses interface.
Defines custom Protocol Buffers types that don't exist as native types in Python.
See also: https://developers.google.com/protocol-buffers/docs/proto3#scalar

`pure-protobuf` contributors © 2011-2019
"""

from typing import NewType

double = NewType('double', float)
fixed32 = NewType('fixed32', int)
fixed64 = NewType('fixed64', int)
sfixed32 = NewType('sfixed32', int)
sfixed64 = NewType('sfixed64', int)
sint32 = NewType('sint32', int)
sint64 = NewType('sint64', int)
uint = NewType('uint', int)  # is not a part of the standard
uint32 = NewType('uint32', int)
uint64 = NewType('uint64', int)
int32 = uint32  # TODO: is it actually the same?
int64 = uint64  # TODO: is it actually the same?

# Not available in `typing`.
NoneType = type(None)



"""
`pure-protobuf` contributors © 2011-2019
"""

from enum import IntEnum


# Wire types.
# See also https://developers.google.com/protocol-buffers/docs/encoding#structure
class WireType(IntEnum):
    VARINT = 0
    LONG_LONG = 1  # 64 bit
    BYTES = 2
    LONG = 5  # 32 bit



"""
`pure-protobuf` contributors © 2011-2019
"""

from abc import ABC, abstractmethod
from io import BytesIO
from typing import Any, BinaryIO, Union

# Type hinting doesn't recognize `BytesIO` as an instance of `BinaryIO`.
IO = Union[BinaryIO, BytesIO]


class Dumps(ABC):
    @abstractmethod
    def dump(self, value: Any, io: IO):
        """
        Serializes a value into a file-like object.
        """
        raise NotImplementedError()

    def dumps(self, value: Any) -> bytes:
        """
        Serializes a value into a byte string
        """
        with BytesIO() as io:
            self.dump(value, io)
            return io.getvalue()


class Loads(ABC):
    @abstractmethod
    def load(self, io: IO) -> Any:
        """
        Deserializes a value from a file-like object.
        """
        raise NotImplementedError()

    def loads(self, bytes_: bytes) -> Any:
        """
        Deserializes a value from a byte string.
        """
        with BytesIO(bytes_) as io:
            return self.load(io)



from setuptools import find_packages, setup

setup(
    name='pure-protobuf',
    version='1.0.0',
    description='Python implementation of Protocol Buffers data types with dataclasses support',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Pavel Perestoronin',
    author_email='eigenein@gmail.com',
    url='https://github.com/eigenein/protobuf',
    packages=find_packages(exclude=['tests']),
    zip_safe=True,
    install_requires=[
        'dataclasses>=0.6,<1.0; python_version >= "3.6" and python_version < "3.7"',
    ],
    extras_require={
        'dev': ['flake8', 'isort', 'pytest', 'pytest-cov', 'coveralls'],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
    ],
)



#!/usr/bin/env python3

"""
Python 3.6+ type hinting interface.

`pure-protobuf` contributors © 2011-2019
"""

from abc import ABC
from collections import abc
from enum import IntEnum
from io import BytesIO
from typing import Any, ClassVar, Dict, Iterable, List, Tuple, Type, TypeVar, Union, get_type_hints

from pure_protobuf import serializers, types
from pure_protobuf.enums import WireType
from pure_protobuf.fields import Field, NonRepeatedField, PackedRepeatedField, UnpackedRepeatedField
from pure_protobuf.io_ import IO
from pure_protobuf.serializers import IntEnumSerializer, MessageSerializer, PackingSerializer, Serializer
from pure_protobuf.types import NoneType

try:
    import dataclasses
except ImportError:
    raise ImportError('dataclasses interface requires dataclasses support')


T = TypeVar('T')


@dataclasses.dataclass
class Message(ABC):
    """
    Virtual base class for a message type.
    See also: https://docs.python.org/3/library/abc.html#abc.ABCMeta.register
    """

    serializer: ClassVar[Serializer]

    def validate(self):
        self.serializer.validate(self)

    def dump(self, io: IO):
        """
        Serializes a message into a file-like object.
        """
        self.validate()
        self.serializer.dump(self, io)

    def dumps(self) -> bytes:
        """
        Serializes a message into a byte string.
        """
        with BytesIO() as io:
            self.dump(io)
            return io.getvalue()

    def merge_from(self: T, other: T):
        """
        Merge another message into the current one, as if with the ``Message::MergeFrom`` method.
        """
        for field_ in self.__protobuf_fields__.values():  # type: Field
            setattr(self, field_.name, field_.merge(
                getattr(self, field_.name),
                getattr(other, field_.name),
            ))


def load(cls: Type[T], io: IO) -> T:
    """
    Deserializes a message from a file-like object.
    """
    return cls.serializer.load(io)


def loads(cls: Type[T], bytes_: bytes) -> T:
    """
    Deserializes a message from a byte string.
    """
    with BytesIO(bytes_) as io:
        return load(cls, io)


def field(number: int, *args, **kwargs) -> dataclasses.Field:
    """
    Convenience method to assign field numbers.
    Calls the standard ``dataclasses.field`` function with the metadata assigned.
    """
    return dataclasses.field(*args, metadata={'number': number}, **kwargs)


def message(cls: Type[T]) -> Type[T]:
    """
    Returns the same class as was passed in, with additional dunder attributes needed for
    serialization and deserialization.
    """

    type_hints = get_type_hints(cls)

    try:
        # Used to list all fields and locate fields by field number.
        cls.__protobuf_fields__: Dict[int, Field] = dict(
            make_field(field_.metadata['number'], field_.name, type_hints[field_.name])
            for field_ in dataclasses.fields(cls)
        )
    except KeyError as e:
        raise TypeError(f'type is not serializable: {e}') from e

    # noinspection PyUnresolvedReferences
    Message.register(cls)
    cls.serializer = MessageSerializer(cls)
    cls.validate = Message.validate
    cls.dump = Message.dump
    cls.dumps = Message.dumps
    cls.merge_from = Message.merge_from
    cls.load = classmethod(load)
    cls.loads = classmethod(loads)

    return cls


def make_field(number: int, name: str, type_: Any) -> Tuple[int, Field]:
    """
    Figure out how to serialize and de-serialize the field.
    Returns the field number and a corresponding ``Field`` instance.
    """
    is_optional, type_ = get_optional(type_)
    is_repeated, type_ = get_repeated(type_)

    if isinstance(type_, type) and issubclass(type_, Message):
        # Embedded message.
        serializer = PackingSerializer(type_.serializer)
    elif isinstance(type_, type) and issubclass(type_, IntEnum):
        # Enumeration.
        # See also: https://developers.google.com/protocol-buffers/docs/proto3#enum
        serializer = IntEnumSerializer(type_)
    else:
        # Predefined type.
        try:
            serializer = SERIALIZERS[type_]
        except KeyError:
            import pure_protobuf.google  # `dataclasses_` has to already be imported beforehand
            serializer = pure_protobuf.google.SERIALIZERS[type_]

    if not is_repeated:
        # Non-repeated field.
        return number, NonRepeatedField(number, name, serializer, is_optional)
    elif serializer.wire_type != WireType.BYTES:
        # Repeated fields of scalar numeric types are packed by default.
        # See also: https://developers.google.com/protocol-buffers/docs/encoding#packed
        return number, PackedRepeatedField(number, name, serializer)
    else:
        # Repeated field of other type.
        return number, UnpackedRepeatedField(number, name, serializer)


def get_optional(type_: Any) -> Tuple[bool, Any]:
    """
    Extracts ``Optional`` type annotation if present.
    This may be useful if a user wants to annotate a field with ``Optional[...]`` and set default to ``None``.
    """
    if getattr(type_, '__origin__', None) is Union:
        args = set(type_.__args__)

        # Check if it's a union of `NoneType` and something else.
        if len(args) == 2 and NoneType in args:
            # Extract inner type.
            type_, = args - {NoneType}
            return True, type_

    return False, type_


def get_repeated(type_: Any) -> Tuple[bool, Any]:
    """
    Extracts ``repeated`` modifier if present.
    """
    if getattr(type_, '__origin__', None) in (list, List, Iterable, abc.Iterable):
        type_, = type_.__args__
        return True, type_
    else:
        return False, type_


SERIALIZERS: Dict[Type, Serializer] = {
    bool: serializers.BooleanSerializer(),
    bytes: serializers.bytes_serializer,
    float: serializers.FloatSerializer(),
    int: serializers.signed_varint_serializer,  # is not a part of the standard
    str: serializers.StringSerializer(),
    types.double: serializers.DoubleSerializer(),
    types.fixed32: serializers.UnsignedFixed32Serializer(),
    types.fixed64: serializers.UnsignedFixed64Serializer(),
    types.sfixed32: serializers.SignedFixed32Serializer(),
    types.sfixed64: serializers.SignedFixed64Serializer(),
    types.sint32: serializers.SignedInt32Serializer(),
    types.sint64: serializers.SignedInt64Serializer(),
    types.uint32: serializers.UnsignedInt32Serializer(),
    types.uint64: serializers.UnsignedInt64Serializer(),
    types.uint: serializers.unsigned_varint_serializer,  # is not a part of the standard
    # TODO: `map`.
}


__all__ = [
    'field',
    'load',
    'loads',
    'message',
    'Message',
]
